/*  shogiopening
 *  Copyright (C) 2000 Sverre Hvammen Johansen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* Some statistics */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include "graph.h"
#include "compute.h"

static int compute_node_number_rek(t_node *n, int i)
{
  GSList *l;
  if (n->mark==0)
    { 
      n->mark=1;
      n->node_number= i++;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  i=compute_node_number_rek(((t_arc *)l->data)->to_node, i);
	}
    }
  return (i);
}

int compute_node_number(t_node *n)
{
  return compute_node_number_rek(n, 1);
}

static int compute_path_rek(t_node *n)
{
  GSList *l;
  int max=0, m;
  if (n->mark==0)
    {
      n->mark=1;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  m= compute_path_rek(((t_arc *)l->data)->to_node);
	  if (m>max) max=m;
	}
      n->mark= max+1;
    }
  return (n->mark);
}

static int compute_path(t_node *n)
{
  int i;
  i= compute_path_rek (n);
  mark_nodes(n,0);
  return (i);
}

static int compute_nodes_rek(t_node *n)
{
  GSList *l;
  int sum=0;
  if (n->mark==0)
    { 
      n->mark=1;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  sum+= compute_nodes_rek(((t_arc *)l->data)->to_node);
	}
      sum= sum+1;
    }
  return (sum);
}

static int compute_nodes(t_node *n)
{
  int i;
  i= compute_nodes_rek(n);
  mark_nodes(n,0);
  return (i);
}

static t_data *find_current_data_rek(t_graph *g, t_node *n)
{
  t_data *ret= NULL;
  if (n->mark==0)
    { 
      n->mark=1;
      if (!n->leaf && n->tos && n->tos->next==NULL)
	ret= find_current_data_rek(g, ((t_arc *)n->tos->data)->to_node);
      else if (!n->leaf && n->tos)
	;
      else if (n->data)
	ret= n->data;
      else
	ret= n->data= new_data(g);
      n->mark=0;
    }
  return ret;
}

t_data *find_current_data(t_graph *g, t_node *n)
{
  return find_current_data_rek(g, n);
}

static int compute_leaves_rek(t_graph *g, t_node *n)
{
  GSList *l;
  int sum=0;
  int plusnode=1;
  if (n->mark==0)
    { 
      n->mark=1;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  sum+= compute_leaves_rek(g, ((t_arc *)l->data)->to_node);
	  plusnode=n->leaf;
	}
      if(plusnode) sum++;
    }
  return (sum);
}

static int compute_leaves(t_graph *g, t_node *n)
{
  int i;
  i= compute_leaves_rek(g, n);
  mark_nodes(n,0);
  return (i);
}

#include "compress_board.h"

t_compute *compute (t_graph *g, t_node *n)
{
  if (!n->compute)
    {
      n->compute= g_chunk_new (t_compute, g->compute);
      n->compute->update_seq= -1;
    }
  if (n->compute->update_seq != g->update_seq)
    {
      n->compute->update_seq= g->update_seq;
      n->compute->pathl= compute_path (n);
      n->compute->nodes= compute_nodes (n);
      n->compute->leaves= compute_leaves (g, n);
      n->compute->current_data= find_current_data (g, n);
    }
  return n->compute;
}
