/*  shogiopening
 *  Copyright (C) 2000 Sverre Hvammen Johansen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "graph.h"
#include "draw.h"
#include "compute.h"
#include "board.h"
#include "support.h"
#include "pieces.h"
#include "move.h"
#include "note.h"
#include "fileio.h"
#include "preferences.h"
#include "compress_board.h"

GtkWidget *app1;
GtkWidget *viewport[10][10];
GtkWidget *pixtable[10][10];
GtkWidget *labeltable[8][10];

t_node *current_node;

char color_xpm_d[12];

char usecol[10][10];
int x1, y1;

gboolean train_mode=FALSE;
gboolean editable=FALSE;
gboolean view_flipped=FALSE;
gboolean auto_promote;
int style;
gboolean style_west, style_clasical;

static void select_popup_entry_to (GtkWidget *window, t_node *n)
{
  move_forward (main_graph,  n);
}

static void select_popup_entry_from (GtkWidget *window, t_node *n)
{
  move_backward (main_graph,  n);
}

static void
flip_if_flipped (int *x, int *y)
{
  if (view_flipped)
    {
      if (HAND (*x, *y))
	{
	  int t;
	  t= *x; *x= *y; *y= t;
	} 
      else
	{
	  *y= 10 - *y;
	  *x= 10 - *x;
	}
    }
}

gboolean
on_viewport_button_press_event         (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  int x, y;
  for (x=0;x<=9;x++)
    for (y=0;y<=9;y++)
      if (widget == viewport[x][y])
	{
	  flip_if_flipped (&x, &y);

	  if (usecol[x][y]==YELLOW && event->button == 3)
	    {
	      static GtkWidget *menu;
	      GSList *l;

	      if (menu)
		gtk_widget_destroy (menu);
	      menu= gtk_menu_new ();
	      for (l=main_graph->current_node->tos; l; l= l->next)
		{
		  t_arc *a;
		  GtkWidget *menu_item;
		  
		  a= (t_arc *) l->data;
		  if (a->from.x!=x || a->from.y!=y) continue;
		  menu_item=gtk_menu_item_new_with_label(a->move);
		  gtk_signal_connect (GTK_OBJECT(menu_item),
				      "activate",
				      GTK_SIGNAL_FUNC(select_popup_entry_to),
				      a->to_node);
		  gtk_menu_append (GTK_MENU (menu), menu_item);
		  gtk_widget_show(menu_item);
		}
	      gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 3, 0);
	      
	      return TRUE;
	    }
	  if (usecol[x][y]==BLUE && event->button == 3)
	    {
	      static GtkWidget *menu;
	      GSList *l;

	      if (menu)
		gtk_widget_destroy (menu);
	      menu= gtk_menu_new ();
	      for (l=main_graph->current_node->froms; l; l= l->next)
		{
		  t_arc *a= (t_arc *) l->data;
		  GtkWidget *menu_item=gtk_menu_item_new_with_label(a->move);
		  gtk_signal_connect (GTK_OBJECT(menu_item),
				      "activate",
				      GTK_SIGNAL_FUNC(select_popup_entry_from),
				      a->from_node);
		  gtk_menu_append (GTK_MENU (menu), menu_item);
		  gtk_widget_show(menu_item);
		}
	      gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 3, 0);
	      
	      return TRUE;
	    }
	  draw_board(main_graph, x, y);
	  return TRUE;
	}
  return TRUE;
}

gboolean
on_viewport_button_release_event         (GtkWidget       *widget,
					  GdkEventButton  *event,
					  gpointer         user_data)
{
  int x, y;
  for (x=0;x<=9;x++)
    for (y=0;y<=9;y++)
      if (widget == viewport[x][y])
	{
	  flip_if_flipped (&x, &y);

	  if ((usecol[x][y]==YELLOW && !(x==x1 && y==y1)) || usecol[x][y]==BLUE)
	    draw_board(main_graph, 0, 0);
	  else if (x1!=0 && (x1!=x || y1!=y))
	    draw_board(main_graph, x, y);
	  return TRUE;
	}
  return TRUE;
}

static void transform_cordinates (int x, int y, int *xx, int *yy)
{
  if (x>=1 && x<=9)
    {
      *xx=x; *yy=y;
    } 
  else
    {
      int piece= 0;
      switch (y)
	{
	case 1:
	case 9:
	  if(x%2) piece=LANCE; else piece=KNIGHT;
	  break;
	case 2:
	case 8:
	  if(x%2) piece=SILVER; else piece=GOLD;
	  break;
	case 3:
	case 7:
	  if(x%2) piece=ROOK; else piece=BISHOP;
	  break;
	case 4:
	case 6:
	  piece=PAWN;
	  break;
	}
      if (x<1) { *xx=0; *yy= piece; } else { *yy= 0; *xx= piece; }
    }
}

void mark_board_position (int x, int y, char mark)
{
  static int xprev, yprev;
  if (x || y)
    {
      xprev= x; yprev=y;
    } 
  else
    {
      x= xprev; y= yprev;
    }	
  if (x || y)
    {
      if (!mark && !(HAND (x, y)))
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport[x][y]),
				      GTK_SHADOW_IN);
      else if (mark && (HAND (x, y)))
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport[x][y]),
				      GTK_SHADOW_OUT);
      else
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport[x][y]),
				      GTK_SHADOW_NONE);
    }
}

static gboolean up (GdkEventKey *event)
{
  switch (event->keyval)
    {
      case GDK_Up:
      case GDK_KP_Home:
      case GDK_KP_7:
      case GDK_KP_Up:
      case GDK_KP_8:
      case GDK_KP_Page_Up:
      case GDK_KP_9:
      return TRUE;
    }
  return FALSE;
}

static gboolean down (GdkEventKey *event)
{
  switch (event->keyval)
    {
      case GDK_Down:
      case GDK_KP_End:
      case GDK_KP_1:
      case GDK_KP_Down:
      case GDK_KP_2:
      case GDK_KP_Page_Down:
      case GDK_KP_3:
      return TRUE;
    }
  return FALSE;
}

static gboolean right (GdkEventKey *event)
{
  switch (event->keyval)
    {
      case GDK_Right:
      case GDK_KP_Page_Up:
      case GDK_KP_9:
      case GDK_KP_Right:
      case GDK_KP_6:
      case GDK_KP_Page_Down:
      case GDK_KP_3:
      return TRUE;
    }
  return FALSE;
}

static gboolean left (GdkEventKey *event)
{
  switch (event->keyval)
    {
      case GDK_Left:
      case GDK_KP_Home:
      case GDK_KP_7:
      case GDK_KP_Left:
      case GDK_KP_4:
      case GDK_KP_End:
      case GDK_KP_1:
      return TRUE;
    }
  return FALSE;
}

static gboolean enter(GdkEventKey *event)
{
  switch (event->keyval)
    {
    case GDK_Return:
    case GDK_space:
    case GDK_KP_0:
    case GDK_KP_5:
    case GDK_KP_Insert:
    case GDK_KP_Enter:
    case GDK_KP_Next:
      return TRUE;
    }
  return FALSE;
}

gboolean
on_app1_key_press_event                (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
  static int x, y;
  static int xx, yy;
  gboolean dorest=FALSE;
  if (x==0 && y==0 && (up(event) || down(event) ||
      left(event) || right(event)))
    x= y= 5;
  if ((x || y) && (up(event) || down(event) ||
      left(event) || right(event)))
    {
      mark_board_position (xx, yy, FALSE);

      if (x>5)
	{
	  if (y>1 && (x!= -1 || y!=7) && (x!=0 || y!=6) &&
	      up(event))
	    y--;
	  if ((x>1 || (x>0 && y>5) || (x> -1 && y>6)) && 
	      right(event))
	    x--;
	} else dorest=TRUE;	  
      if (y<9 && (x!=10 || y!=4) && (x!=11 || y!=3) &&
	  down(event))
	y++;
      if ((x<9 || (x<10 && y<5) || (x<11 && y<4)) && 
	  left(event))
	x++;
      if (dorest)
	{
	  if (y>1 && (x!= -1 || y!=7) && (x!=0 || y!=6) &&
	      up(event))
	    y--;
	  if ((x>1 || (x>0 && y>5) || (x> -1 && y>6)) && 
	      right(event))
	    x--;
	}	  

      transform_cordinates (x, y, &xx, &yy);

      mark_board_position (xx, yy, TRUE);

    }
  else if ((x || y) && enter(event))
    {
      int xxx= xx, yyy= yy;
      flip_if_flipped (&xxx, &yyy);

      draw_board(main_graph, xxx, yyy);
    }
  return FALSE;
}


static void 
create_view(GtkWidget *widget, int i, int j, int x, int y, int left)
{
  GtkWidget *p;
  viewport[i][j] = gtk_viewport_new (NULL, NULL);
  gtk_widget_show (viewport[i][j]);
  gtk_table_attach (GTK_TABLE (widget), viewport[i][j], x, x+1, y, y+1,
		    (GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
  gtk_widget_set_events (viewport[i][j], 
			 GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
  gtk_signal_connect (GTK_OBJECT (viewport[i][j]), "button_press_event",
                      GTK_SIGNAL_FUNC (on_viewport_button_press_event),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (viewport[i][j]), "button_release_event",
                      GTK_SIGNAL_FUNC (on_viewport_button_release_event),
                      NULL);
  if (HAND (i, j))
    gtk_viewport_set_shadow_type (GTK_VIEWPORT (viewport[i][j]), 
				  GTK_SHADOW_NONE);
  strcpy (color_xpm_d, COLOR_NONE);
  p= gnome_pixmap_new_from_xpm_d(pieces_xpm_d[style][0]);
  gtk_widget_show (p);
  gtk_container_add (GTK_CONTAINER (viewport[i][j]), p);
  pixtable[i][j]= p;

  if (HAND (i, j))
    {
      labeltable[i][j]= gtk_label_new (" ");
      gtk_widget_show (labeltable[i][j]);
      gtk_table_attach (GTK_TABLE (widget), labeltable[i][j], x-left*2+1, x+1-left*2+1, y, y+1,
			(GtkAttachOptions) GTK_EXPAND | GTK_FILL, (GtkAttachOptions) GTK_EXPAND | GTK_FILL, 0, 0);
      gtk_misc_set_alignment (GTK_MISC (labeltable[i][j]), left, 0.5);
    }
}

void
create_board                      (GtkWidget       *widget)
{
  int i,j;
  for (i=1;i<=9;i++)
    for (j=1;j<=9;j++)
      create_view(widget, i, j, 14-i, j, 0);
  create_view(widget, WHITE_HAND_X(PAWN), WHITE_HAND_Y(PAWN), 2, 4, 0);
  create_view(widget, WHITE_HAND_X(ROOK), WHITE_HAND_Y(ROOK), 1, 3, 1);
  create_view(widget, WHITE_HAND_X(BISHOP), WHITE_HAND_Y(BISHOP), 2, 3, 0);
  create_view(widget, WHITE_HAND_X(LANCE), WHITE_HAND_Y(LANCE), 1, 1, 1);
  create_view(widget, WHITE_HAND_X(KNIGHT), WHITE_HAND_Y(KNIGHT), 2, 1, 0);
  create_view(widget, WHITE_HAND_X(SILVER), WHITE_HAND_Y(SILVER), 1, 2, 1);
  create_view(widget, WHITE_HAND_X(GOLD), WHITE_HAND_Y(GOLD), 2, 2, 0);

  create_view(widget, BLACK_HAND_X(PAWN), BLACK_HAND_Y(PAWN), 16, 6, 1);
  create_view(widget, BLACK_HAND_X(ROOK), BLACK_HAND_Y(ROOK), 17, 7, 0);
  create_view(widget, BLACK_HAND_X(BISHOP), BLACK_HAND_Y(BISHOP), 16, 7, 1);
  create_view(widget, BLACK_HAND_X(LANCE), BLACK_HAND_Y(LANCE), 17, 9, 0);
  create_view(widget, BLACK_HAND_X(KNIGHT), BLACK_HAND_Y(KNIGHT), 16, 9, 1);
  create_view(widget, BLACK_HAND_X(SILVER), BLACK_HAND_Y(SILVER), 17, 8, 0);
  create_view(widget, BLACK_HAND_X(GOLD), BLACK_HAND_Y(GOLD), 16, 8, 1);

}


static void flip_viewx()
{
  char s[100];
  char *marker;
  int i;
  GtkWidget *w;
  for (i=1;i<=36;i++)
    {
      sprintf(s, "marker%d", i);
      gtk_label_get(GTK_LABEL(w=get_widget (app1, s)), &marker);
      strcpy (s, marker);
      if (s[0]>='1' && s[0]<='9') s[0]= '1' + 8 - (s[0]-'1');
      if (s[0]>='a' && s[0]<='i') s[0]= 'a' + 8 - (s[0]-'a');
      gtk_label_set_text(GTK_LABEL(w), s);
    }

  view_flipped= !view_flipped;

  gtk_label_set_text (GTK_LABEL(lookup_widget (app1, "down_player")), 
		      view_flipped?_("Sente"):_("Gote"));
  gtk_label_set_text (GTK_LABEL(lookup_widget (app1, "up_player")), 
		      view_flipped?_("Gote"):_("Sente"));
}

void flip_view(t_graph *g)
{
  flip_viewx();
  draw_board(g, 0, 0);

}

static void
on_menu_items_activate_forward         (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  move_forward (main_graph, (t_node *) user_data);
}

static void
on_menu_items_activate_backward        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  move_backward (main_graph, (t_node *) user_data);
}

void delete_menu_item (gpointer data, gpointer user_data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
set_menu(GSList *l, GtkWidget *parent, gboolean forward)
{
  t_arc *a;
  GtkWidget *menu_item;
#if 0
  GtkWidget *menu;
  menu= gtk_menu_new ();
  
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(parent), menu);

  gtk_container_foreach(GTK_CONTAINER (menu), 
			GTK_SIGNAL_FUNC (gtk_widget_destroy), 
			NULL);

  for (; l!= NULL; l=l->next)
    {
      a= (t_arc *) l->data;

      menu_item=gtk_menu_item_new_with_label(a->move);

      gtk_menu_append (GTK_MENU (menu), menu_item);

      if (forward)
	gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			    GTK_SIGNAL_FUNC (on_menu_items_activate_forward),
			    (gpointer) a->to_node);
      else
	gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			    GTK_SIGNAL_FUNC (on_menu_items_activate_backward),
			    (gpointer) a->from_node);
	
      gtk_widget_show(menu_item);
    }
#else
  GList *ldestroy;
  while ((ldestroy = GTK_MENU_SHELL(GTK_MENU_ITEM(parent)->submenu)
	  ->children->next->next->next))
    gtk_widget_destroy(GTK_WIDGET(ldestroy->data));

  for (; l!= NULL; l=l->next)
    {
      a= (t_arc *) l->data;

      menu_item=gtk_menu_item_new_with_label(a->move);

      gtk_menu_append (GTK_MENU (GTK_MENU_ITEM(parent)->submenu), menu_item);

      if (forward)
	gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			    GTK_SIGNAL_FUNC (on_menu_items_activate_forward),
			    (gpointer) a->to_node);
      else
	gtk_signal_connect (GTK_OBJECT (menu_item), "activate",
			    GTK_SIGNAL_FUNC (on_menu_items_activate_backward),
			    (gpointer) a->from_node);
	
      gtk_widget_show(menu_item);
    }
  
#endif
}

static void draw_piece(int x, int y, int piece, int color)
{

  static char a_style[10][10], a_color[10][10], a_piece[10][10];
  int num;

  if (x==0 && y==0) return;
  if (HAND (x, y))
    {
      char a[3];
      if (PIECE_ON_HAND (x, y) > 7) return;
      
      num= piece;
      if (num) 
	piece= (PIECE_ON_HAND(x,y))|((WHITE_HAND(x, y))?WHITE:BLACK);
      else piece= 0;

      flip_if_flipped (&x, &y);

      if (num>1) sprintf(a,"%d",num); else sprintf(a," ");
      gtk_label_set_text(GTK_LABEL(labeltable[x][y]), a);
    }
  else
    {
      flip_if_flipped (&x, &y);
    }
  if(train_mode && color!=BLUE) color=0;
  if(view_flipped && piece) piece= piece^BLACK;

  if (a_style[x][y] != style || a_color[x][y] != color ||
      a_piece[x][y] != piece)
    {
      {
	if(!color)
	  {
	    if(piece)
	      strcpy (color_xpm_d, COLOR_BLACK);
	    else
	      strcpy (color_xpm_d, COLOR_NONE);
	  }
	else
	  strcpy (color_xpm_d, color_xpm_d_table[color]);
      }
      
      gnome_pixmap_load_xpm_d( GNOME_PIXMAP(pixtable[x][y]), 
			       pieces_xpm_d[style][piece]);
      a_style[x][y]= style;
      a_color[x][y]= color;
      a_piece[x][y]= piece;
    }
}

static void draw_statistics(t_graph *g, int leafssum, int nodessum, int pathlength, 
			    t_node *n, int turn)
{
  t_arc *a=NULL;
  char s[30];

  if (leafssum>0)
    sprintf(s,"%d", leafssum);
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "leaves")), s);

  if (nodessum)
    sprintf(s,"%d", nodessum);
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "nodes")), s);

  if (pathlength)
    sprintf(s,"%d", pathlength);
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "toend")), s);

  if (n && move_number(n)>0)
    sprintf(s,"%d", move_number(n));
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "move_num")), s);

  if (n)
    sprintf(s,"%d", g_slist_length (n->froms));
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "pred")), s);

  if (n)
    sprintf(s,"%d", g_slist_length (n->tos));
  else
    s[0]='\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "suc")), s);

  if (n && n->froms)
    {
      t_arc *a2;
      GSList *l;
      a= (t_arc *) n->froms->data;
      
      for (l=n->froms; l!= NULL; l=l->next)
	{
	  a2= (t_arc *) l->data;
	  if (a2->from_node==n->prev) 
	    {
	      a=a2;
	      break;
	    }
	}

      if (move_number(n)<=0)
	{
	  if (move_number(n)%2)
	    sprintf(s,"    :  %s", n?n->prev?a->move:"":"");
	  else
	    sprintf(s,"    :   . . .   %s", n?n->prev?a->move:"":"");
	}
      else
      if (move_number(n)%2)
	sprintf(s,"%d:  %s", (int) ((move_number(n)+1)/2), 
		n?n->prev?a->move:"":"");
      else
	sprintf(s,"%d:   . . .   %s", (int) ((move_number(n)+1)/2), 
		n?n->prev?a->move:"":"");
    } else s[0]= '\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "move")), s);

  if (n==NULL || turn==BLACK)
    strcpy (s, _("Sentes turn"));
  else
    strcpy (s, _("Gotes turn"));
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "turn")), s);

  if (a && a->bad)
    strcpy (s, _("Bad move"));
  else
    s[0]= '\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "mark_bad")), s);
 
  if (n && n->leaf)
    strcpy (s, _("Leaf"));
  else
    s[0]= '\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "mark_leaf")), s);

  if (n && n->note && *n->note)
    strcpy (s, _("Note available"));
  else
    s[0]= '\0';
  gtk_label_set_text(GTK_LABEL(lookup_widget(app1, "note")), s);

  display_note(g, n);
}

void draw_board(t_graph *g, int x2, int y2)
{
  int x,y;
  t_arc *a;
  GSList *l;
  char board[10][10];
  if (g && g->current_node)
#ifdef COMPRESS_BOARD
    uncompress_board (board, g->current_node->board);
#else
  copy_board (board, g->current_node->board);
#endif

  if(x2==0 && y2== 0)
    {
      x1= y1= 0;
    } else
  if(usecol[x2][y2]==YELLOW)
    {
      int possible_moves;
      t_node *auto_move_to= NULL;
      if (auto_move_on) possible_moves=0; else possible_moves=2;
      for (x=0;x<=9;x++)
	for (y=0;y<=9;y++)
	  if (usecol[x][y]==GREEN || usecol[x][y]==RED)
	    draw_piece(x, y, board[x][y], usecol[x][y]=0);
      x1= x2; y1= y2;
      for (l=g->current_node->tos; l!=NULL;l=l->next)
	{
	  a= (t_arc *)l->data;
	  if (a->from.x==x1 && a->from.y==y1)
	    {
	      draw_piece(a->to.x, a->to.y, board[a->to.x][a->to.y], 
			 usecol[a->to.x][a->to.y]=(a->bad?RED:GREEN));
	      if (a->bad) possible_moves=2; else possible_moves++;
	      auto_move_to= a->to_node;
	    }
	}
      if (possible_moves==1)
	{
	  move_forward (g, auto_move_to);
	}
      else
	{
	  int leafssum=0;
	  int nodessum=0;
	  int pathlength=0;
	  for (l=g->current_node->tos; l!=NULL;l=l->next)
	    {
	      t_compute *com;
	      a= (t_arc *)l->data;
	      if (a->from.x==x1 && a->from.y==y1)
		{
		  com= compute (g, a->to_node);
		  leafssum+= com->leaves;
		  nodessum+= com->nodes;
		  if (com->pathl > pathlength) pathlength= com->pathl;
		}
	    }
	  draw_statistics(g, leafssum, nodessum, 
			  pathlength, g->current_node, TURN (board));
	}
      return;
    } else
  if(usecol[x2][y2]==GREEN || usecol[x2][y2]==RED)
    {
      for (l=g->current_node->tos; l!=NULL;l=l->next)
	{
	  a= (t_arc *)l->data;
	  if (a->from.x==x1 && a->from.y==y1 &&
	      a->to.x==x2 && a->to.y==y2)
	    {
	      move_forward (g, a->to_node);
	      return;
	    }
	}
      x1= y1= 0;
    } else
  if(editable && board[x1][y1] &&
     !illegal_move(x1, y1, x2, y2, which_piece(board[x1][y1]), board))
    {
      g->current_node= 
	add_one_node (g, g->current_node, 
		      x1, y1, x2, y2, auto_promote, NULL, board);
      if (auto_flip_on) flip_view(g);
      else
	draw_board(g, 0, 0);
      return;
    } else
  if(usecol[x2][y2]==BLUE)
    {
      move_backward (g, g->current_node->prev);
      return;
    } else
  if(editable)
    {
      x1= x2; y1= y2;
    }

  if (g && g->current_node)
    { 
      set_menu(g->current_node->tos, 
	       lookup_widget (app1, "Tos"), 
	       TRUE);
      set_menu(g->current_node->froms,  
	       lookup_widget (app1, "Froms"), 
	       FALSE);
    }
  else
    {
      set_menu(NULL, 
	       lookup_widget (app1, "Tos"), 
	       TRUE);
      set_menu(NULL,
	       lookup_widget (app1, "Froms"), 
	       FALSE);
    }

  if (g && g->current_node)
    {
      t_compute *com;
      com= compute (g, g->current_node);
      draw_statistics(g, com->leaves, com->nodes, com->pathl, g->current_node, TURN(board));
    }
  else
    draw_statistics(g, 0, 0, 0, NULL, 0);

  for (x=0;x<=9;x++)
    for (y=0;y<=9;y++)
      usecol[x][y]=0;

  if (g && g->current_node)
    {
      for (l=g->current_node->tos; l!=NULL;l=l->next)
	{
	  a= (t_arc *)l->data;
	  usecol[a->from.x][a->from.y]=YELLOW;
	}
      
      for (l=g->current_node->froms; l!=NULL;l=l->next)
	{
	  a= (t_arc *)l->data;
	  if (a->from_node==g->current_node->prev)
	    usecol[a->to.x][a->to.y]=BLUE;
	}

      for (x=0;x<=9;x++)
	for (y=0;y<=9;y++)
	  draw_piece(x, y, board[x][y], usecol[x][y]);
    }
  else
    for (x=0;x<=9;x++)
      for (y=0;y<=9;y++)
	draw_piece(x, y, 0, 0);
}

void init_draw ()
{
  create_board (get_widget (app1, "table2")); 
}
