/*  shogiopening
 *  Copyright (C) 2000 Sverre Hvammen Johansen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <stdio.h>
#include "graph.h"
#include "hash.h"
#include "board.h"
#include "compute.h"
#include "ext.h"
#include "ask.h"
#include "fileio.h"
#include "compress_board.h"

/* Data */

int yyline;

t_data *new_data (t_graph *g)
{
  return (g_chunk_new0(t_data, g->data));
}

/* Arc */

t_arc *new_arc(t_graph *g, char *move, t_node *node)
{
  t_arc *a= g_chunk_new0(t_arc, g->arc);
  a->move= move;
  a->to_node= node;
  node->froms= g_slist_prepend(node->froms, a);
  a->line= yyline;
  return (a);
}

static t_arc *copy_arc(t_graph *g, t_arc *a1, 
		       t_node *from_node, t_node *to_node)
{
  t_arc *a2;
  a2= new_arc (g, a1->move, to_node);
  from_node->tos= g_slist_prepend(from_node->tos, a2);
  a2->from_node= from_node;
  a2->from=a1->from;
  a2->to=a1->to;
  a2->extra=a1->extra;
  return (a2);
}

/* Node */

static int move_number_rek (t_node *n, int limit)
{
  if (limit<=1) return -1000;
  if (n->prev==NULL && n->froms)
    n->prev= ((t_arc *)n->froms->data)->from_node;
  if (n->prev) return (move_number_rek (n->prev, limit -1)+1);
  return (0);
}

int move_number (t_node *n)
{
  return move_number_rek (n, 999);
}

t_node *new_node(t_graph *g, long number, char *note)
{
  t_node *n= g_chunk_new0 (t_node, g->node);
  static int counter;
  n->node_number= number;
  n->note= note;
  if ((counter++ & 2<<9) == 0) update_interface (g);
  return n;
}

void mark_nodes(t_node *n, int m)
{
  GSList *l;
  if (n->mark!=m)
    {
      n->mark=m;
      for (l=n->tos; l!=NULL; l=l->next)
	mark_nodes(((t_arc *)l->data)->to_node, m);
    }
}

void free_node (t_graph *g, t_node *n)
{
  GSList *l;
  t_arc *a;
  static int counter;
  t_node *nx;
  g_slist_push_allocator(g->gslist);
  if (n->data)
    {
      if (n->froms && n->froms->next==NULL && 
	  ((t_arc *)n->froms->data)->from_node &&
	  ((t_arc *)n->froms->data)->from_node->data==NULL)
	((t_arc *)n->froms->data)->from_node->data= n->data;
      else
	g_chunk_free(n->data, g->data);
    }

  for (l=n->froms; l!=NULL; l=l->next)
    {
      a=(t_arc *)l->data;
      if (a->from_node)
	{
	  nx= a->from_node;
	  nx->tos= g_slist_remove (nx->tos, a);
	  if (nx->next==n)nx->next=NULL;
	}
      g_chunk_free(a, g->arc);
    }
  for (l=n->tos; l!=NULL; l=l->next)
    {
      a=(t_arc *)l->data;
      nx= a->to_node;
      nx->froms= g_slist_remove (nx->froms, a);
      if (nx->prev==n)nx->prev=NULL;
      g_chunk_free(a, g->arc);
    }
  g_slist_free(n->tos);
  g_slist_free(n->froms);
  delete_node_in_hash_table(n);
  g_chunk_free(n, g->node);
  g_slist_pop_allocator();
  if ((counter++ & 2<<12) == 0)update_interface(NULL);
}

#if 0
static char free_nodes_rek(t_graph *g, t_node *n)
{
  GSList *l;

  if (n->mark==0)
    {
      n->mark=1;

      /* Trikky part. NB free_nodes_rek if returned TRUE will update 
         n->tos. It may even remove several entries. Therefor we must
         start iterating from the start of the list each time
         free_nodes_rek returns TRUE */
      while (TRUE)
	{
	  for (l=n->tos; l!=NULL; l=l->next)
	    if (free_nodes_rek(g, ((t_arc *)l->data)->to_node))
	      goto next;
	  break;
	next:
	}

      free_node(g, n);
      return TRUE;
    }
  return FALSE;
}

void free_nodes(t_graph *g, t_node *n)
{
  if (n) free_nodes_rek(g, n);
}
#endif

static void add_from_next_nodes_rek (t_node *n)
{
  GSList *l;
  t_arc *a;
  if (n->mark==0)
    { 
      n->mark=1;
      if (n->tos && n->tos->next==NULL) 
	n->next= ((t_arc *)n->tos->data)->to_node;
      else
	n->next= NULL;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  a= (t_arc *)l->data;
	  add_from_next_nodes_rek(a->to_node);
	  a->from_node= n;
	}
    }
}

void add_from_next_nodes (t_node *n)
{
  add_from_next_nodes_rek (n);
  mark_nodes (n, 0);
}

/* Merge and fold graphs */

t_node *add_one_node (t_graph *g, t_node *n, int x1, int y1, int x2, int y2,
		     char auto_promote, char *move, char board[10][10])
{
  t_node *nx;
  t_arc *a;
#ifdef COMPRESS_BOARD
  int cboard2[9];
#endif
  char board2[10][10];

  t_cord from, to, extra;

  g->unsaved= 1;
  g->update_seq++;
  
  if (move==NULL) move=text_move(x1, y1, x2, y2, 
				 which_piece(board[x1][y1]), 
				 board, auto_promote);

  g_slist_push_allocator(g->gslist);

  move_board(board, board2, move, &from, &to, &extra);
#ifdef COMPRESS_BOARD
  compress_board (board2, cboard2);
  nx= find_node_in_hash_table (g, cboard2, NULL);
#else
  nx= find_node_in_hash_table (g, board2, NULL);
#endif  
  a= new_arc(g, move, nx);
  a->from= from;
  a->to= to;
  a->extra= extra;
  
  n->tos= g_slist_prepend(n->tos, a);
  a->from_node= n;
  
  nx->prev= n;
  n->next= nx;

  g_slist_pop_allocator();
  return nx;
}


static void add_next_nodes_rek (t_node *n)
{
  GSList *l;
  static int counter;
  if (n->mark==0)
    {
      n->mark= 1;
      if (n->tos && n->tos->next==NULL) 
	n->next= ((t_arc *)n->tos->data)->to_node;
      else
	n->next= NULL;
      for (l=n->tos; l!=NULL; l=l->next)
	{
	  add_next_nodes_rek(((t_arc *)l->data)->to_node);

	}
      if ((counter++ & 2<<14) == 0)update_interface(NULL);
    }
}

static void add_next_nodes (t_node *n)
{
  add_next_nodes_rek (n);
  mark_nodes (n, 0);
}

t_graph *new_graph ()
{
  t_graph *g;
  g= (t_graph *) g_malloc0 (sizeof(t_graph));
  g->node= g_mem_chunk_create(t_node, 128, G_ALLOC_AND_FREE);
  g->arc= g_mem_chunk_create(t_arc, 128, G_ALLOC_AND_FREE);
  g->data= g_mem_chunk_create(t_data, 128, G_ALLOC_AND_FREE);
  g->compute= g_mem_chunk_create(t_compute, 128, G_ALLOC_AND_FREE);
  g->gslist= g_allocator_new("gslist", 128);
  g->main= NULL;

  return g;
}

void free_graph (t_graph *g)
{
#if 0
  mark_nodes(g->main, 0);
  if (g->main) free_nodes (g, g->main);
#endif
  g_mem_chunk_destroy (g->node);
  g_mem_chunk_destroy (g->arc);
  g_mem_chunk_destroy (g->data);
  g_mem_chunk_destroy (g->compute);
  g_allocator_free (g->gslist);
  g_free (g);
}

t_graph *new_main_graph ()
{
  t_graph *g;
#ifdef COMPRESS_BOARD
  char board[10][10];
#endif

  g= new_graph();
  g_slist_push_allocator(g->gslist);
  g->main = new_node (g, 0, NULL);
#ifdef COMPRESS_BOARD
  init_board (board);
  compress_board (board, g->main->board);
#else
  init_board (g->main->board);
#endif
  insert_nodes_in_hash_table(g, g->main);
  g_slist_pop_allocator();
  return g;
}

static void merge_graphs_req (t_graph *g, t_node *n1, t_node *n2)
{
  GSList *l1, *l2;
  t_node *n2x;
  t_arc *a2;
  static int counter;

  if (n2->mark==0)
    {
      n2->mark=1;
      if(n2->leaf)n1->leaf= n2->leaf;
      if(n2->data && n1->data==NULL)
	{
	  n1->data=new_data (g);
	  *n1->data= *n2->data;
	}
      for (l2= n2->tos; l2!=NULL; l2=l2->next)
	{
	  a2= (t_arc *)l2->data;
#if TREE
	  n2x= a2->to_node;
	  for (l1= n1->tos; l1!=NULL; l1=l1->next)
	    {
	      if (boards_is_equal(((t_arc *)l1->data)->to_node->board, 
				  n2x->board)) goto next;
	    }
	  n2x= insert_node_in_hash_table (g, a2->to_node->board, a2->to_node->note);
	  
#else
	  n2x= find_node_in_hash_table(g, a2->to_node->board,
				       a2->to_node->note);
	  for (l1= n1->tos; l1!=NULL; l1=l1->next)
	    {
	      if (((t_arc *)l1->data)->to_node == n2x) goto next;
	    }
#endif
	  copy_arc(g, a2, n1, n2x);
	next:
	  merge_graphs_req(g, n2x, a2->to_node);
	}
      if ((counter++ & 2<<11) == 0)update_interface(NULL);
    }
}

t_graph *merge_graphs (t_graph *g1, t_graph *g2)
{
  /* slist_allocator stack: g1 */
  g_slist_push_allocator(g1->gslist);
  merge_graphs_req (g1, g1->main, g2->main);
  g_slist_pop_allocator();
  add_next_nodes (g1->main);
  free_graph (g2);
  return (g1);
}

static t_graph *fold_graph (t_graph *g2)
{
  return merge_graphs (new_main_graph (), g2);
}

/* PRUNE */

static char prune_graph_rek(t_graph *g, t_node *n, int branch_dist, 
			    int max_dist)
{
  GSList *l, *lnext;
  char all_delete= TRUE;
  if (n->mark==1) return(FALSE);
  n->mark= 1;
  if(n->tos && n->tos->next) branch_dist= 0; else branch_dist++;
  for (l=n->tos; l!=NULL; l=lnext)
    {
      lnext= l->next;
      if (!prune_graph_rek (g, ((t_arc *)l->data)->to_node, branch_dist,
			    max_dist))
	all_delete= FALSE;
    }

  if (all_delete && branch_dist>max_dist)
    {
      free_node (g, n);
      return TRUE;
    }
  return FALSE;
}

void prune_graph(t_graph *g, t_node *n, int max_dist)
{
  g->unsaved= 1;
  prune_graph_rek(g, n, 0, max_dist);
  mark_nodes (g->main, 0);
}

/* Open and Save The graph */

t_graph *open_graph(t_graph *main, char *name, char free)
{
  t_graph *new;
  gboolean fold;

  if (free) 
    {
      free_graph (main);
      main= NULL;
    }

  new=build_graph(name, &fold);

  if (new) new=compute_boards (new);
  if (new)
    {
      add_from_next_nodes (new->main);
      if (main)
	{
	  main= merge_graphs (main, new);
	}
      else if (fold)
	{
	  /* slist_allocator stack: new */
	  main= fold_graph (new);
	}
      else
	{
	  g_slist_push_allocator(new->gslist);
	  insert_nodes_in_hash_table (new, new->main);
	  g_slist_pop_allocator();
	  main= check_duplicate_nodes (new);
	}
    }

  if (main) return (main);
  else
    return (new_main_graph ());
}


static void save_node(FILE *f, t_node *n);
static void save_arc(FILE *f, t_arc *a)
{
  fprintf(f, "(\"%s\" ", a->move);
  if (a->bad) fprintf(f, ":BAD ");
  save_node(f, a->to_node);
  fprintf(f, ")");
}

static void save_node(FILE *f, t_node *n)
{
  GSList *l;
  if (n->mark)
    fprintf(f, "#%d#", n->node_number);
  else
    {
      n->mark=1;
      fprintf(f, "\n#%d=#S(NODE ",n->node_number);
      if (n->note)
	fprintf(f, ":NOTE \"%s\" ",n->note);
      if (n->leaf && n->tos)
	fprintf(f, ":LEAF ");
      if (n->data)
	{
	  if(n->data->sente)
	    fprintf(f, ":SENTE \"%s\" ",n->data->sente);
	  if(n->data->gote)
	    fprintf(f, ":GOTE \"%s\" ",n->data->gote);
	  if(n->data->date)
	    fprintf(f, ":DATE \"%s\" ",n->data->date);
	  if(n->data->event)
	    fprintf(f, ":EVENT \"%s\" ",n->data->event);
	  if(n->data->opening)
	    fprintf(f, ":OPENING \"%s\" ",n->data->opening);
	  if(n->data->result)
	    fprintf(f, ":RESULT \"%s\" ",n->data->result);
	}
#if 0
      fprintf(f, ":FROMS (")
      for (l=n->froms; l!=NULL; l=l->next)
	save_arc(f, (t_arc *)l->data);
      fprintf(f, ") ");
#endif
      fprintf(f, ":TOS (");
      for (l=n->tos; l!=NULL; l=l->next)
	save_arc(f, (t_arc *)l->data);
      fprintf(f, "))");
    }
}

void save_graph(char *name, t_node *n)
{
  FILE *f;
  if (n==NULL) return;

  compute_node_number(n);
  mark_nodes(n,0);

  f= fopen (name, "w");
  if (f==NULL)
    {
      char s[5000];
      sprintf(s, _("Could not write %s"), name);
      display_error(s);
    } 
  else
    {
      save_node(f, n);
      mark_nodes(n,0);
      fclose(f);
    }
}


void save_path(char *name, t_node *n)
{
  FILE *f;
  if (n==NULL) return;

  f= fopen (name, "w");
  if (f==NULL)
    {
      char s[5000];
      sprintf(s, _("Could not write %s"), name);
      display_error(s);
    } 
  else
    {
      t_node *nx;
      int i=0;

      fprintf(f, "[Comment \"WARNING: This is one path in a graph written by Shogiopening and this particular path may not have been played at all.\"]\n");
      for (nx= n; nx; nx=nx->next)
	{
	  if (i++ >999)
	    {
	      display_error(_("Loop in path"));
	      fclose(f);
	      return;
	    }
	  if (nx->data)
	    {
	      if (nx->data->sente && *nx->data->sente)
		fprintf(f, "[Sente \"%s\"]\n", nx->data->sente);
	      if (nx->data->gote && *nx->data->gote)
		fprintf(f, "[Gote \"%s\"]\n", nx->data->gote);
	      if (nx->data->date && *nx->data->date)
		fprintf(f, "[Date \"%s\"]\n", nx->data->date);
	      if (nx->data->event && *nx->data->event)
		fprintf(f, "[Event \"%s\"]\n", nx->data->event);
	      if (nx->data->opening && *nx->data->opening)
		fprintf(f, "[Opening \"%s\"]\n", nx->data->opening);
	      if (nx->data->result && *nx->data->result)
		fprintf(f, "[Result \"%s\"]\n", nx->data->result);
	      break;
	    }
	}
      for (nx= n; nx->next; nx= nx->next)
	{
	  GSList *l;
	  t_arc *a=NULL;
	  for (l=nx->tos; l!=NULL; l=l->next)
	    {
	      a= ((t_arc *)l->data);
	      if (a->to_node == nx) break;
	    }
	  fprintf (f, "%s ", a->move);
	  if (nx->next->note && *nx->next->note)
	    fprintf (f, "{%s} ", nx->next->note);
	  fprintf(f, "\n");
	}
      fprintf(f, "\n");
      fclose(f);
    }
}




